--========================================================--
--                Enhance Battlefield Minimap             --
--                                                        --
-- Author      :  kurapica125@outlook.com                 --
-- Create Date :  2017/03/04                              --
--========================================================--

--========================================================--
Scorpio            "EnhanceBattlefieldMinimap"       "1.1.1"
--========================================================--

max                     = math.max
min                     = math.min
abs                     = math.abs

BattlefieldMinimapPOI   = false

MOVE_MINIMAP_THREADIDX  = 0
ENTER_TASK_ID           = 0
HiddenTime              = 0
MOUSEDOWN_TIME          = 0

----------------------------------------------
-------------- Addon Event Handler -----------
----------------------------------------------
function OnLoad(self)
    _SVDB = SVManager.SVCharManager("EnhanceBattlefieldMinimap_DB")

    _SVDB:SetDefault {
        -- Display Status
        MaskScale       = 1.0,
        MinimapScale    = 1.0,
        VerticalScroll  = 0,
        HorizontalScroll= 0,
        HiddenWhenQuit  = false,

        -- Settings
        HideInCombat    = true,
        LockOnPlayer    = true,
        ScaleModifier   = 1.0,
    }
end

__Thread__()
function OnEnable(self)
    if not IsAddOnLoaded("Blizzard_BattlefieldMinimap") and not _SVDB.HiddenWhenQuit then
        BattlefieldMinimap_LoadUI()
    end

    if not IsAddOnLoaded("Blizzard_BattlefieldMinimap") then
        while Event("ADDON_LOADED") ~= "Blizzard_BattlefieldMinimap" do end
    end

    -- Modify the BattlefieldMinimap
    BattlefieldMinimapBackground:Hide()
    BattlefieldMinimapCloseButton:Hide()
    BattlefieldMinimapCorner:Hide()

    BattlefieldMinimapTab:SetFrameStrata("HIGH")
    BattlefieldMinimapTabLeft:SetTexture(nil)
    BattlefieldMinimapTabMiddle:SetTexture(nil)
    BattlefieldMinimapTabRight:SetTexture(nil)
    BattlefieldMinimapTabText:Hide()

    BattlefieldMinimap:EnableMouseWheel(true)
    BattlefieldMinimap:SetMovable(true)

    -- Fix the BattlefieldMinimap's w/h ratio
    local worldMapWidth, worldMapHeight = WorldMapDetailFrame:GetSize()
    local tileSize = BattlefieldMinimap1:GetWidth()
    local worldTileSize = WorldMapDetailTile1:GetWidth()

    BattlefieldMinimap:SetSize(worldMapWidth / worldTileSize * tileSize, worldMapHeight / worldTileSize * tileSize)

    BattlefieldMinimapScroll = CreateFrame("ScrollFrame", "BattlefieldMinimapScroll", UIParent)

    for i = 1, BattlefieldMinimap:GetNumPoints() do
        BattlefieldMinimapScroll:SetPoint(BattlefieldMinimap:GetPoint(i))
    end

    ORIGIN_WIDTH, ORIGIN_HEIGHT = BattlefieldMinimap:GetSize()

    BattlefieldMinimapScroll:SetSize(ORIGIN_WIDTH * _SVDB.MaskScale, ORIGIN_HEIGHT * _SVDB.MaskScale)
    BattlefieldMinimapScroll:SetScrollChild(BattlefieldMinimap)

    BattlefieldMinimap:ClearAllPoints()
    BattlefieldMinimap:SetPoint("TOPLEFT", BattlefieldMinimapScroll, "TOPLEFT")
    BattlefieldMinimap:SetScale(_SVDB.MinimapScale)

    BattlefieldMinimapScroll:UpdateScrollChildRect()
    BattlefieldMinimapScroll:SetVerticalScroll(_SVDB.VerticalScroll)
    BattlefieldMinimapScroll:SetHorizontalScroll(_SVDB.HorizontalScroll)

    UpdateHitRectInsets()

    BattlefieldMinimap:HookScript("OnMouseWheel", OnMouseWheel)
    BattlefieldMinimap:HookScript("OnMouseDown", OnMouseDown)
    BattlefieldMinimap:HookScript("OnMouseUp", OnMouseUp)
    BattlefieldMinimap:HookScript("OnShow", OnShow)
    BattlefieldMinimap:HookScript("OnHide", OnHide)
    BattlefieldMinimap:HookScript("OnEnter", OnEnter)
    BattlefieldMinimapTab:HookScript("OnEnter", TabOnEnter)
    BattlefieldMinimapTab:HookScript("OnLeave", TabOnLeave)

    -- Don't change scale, it'd blink
    BattlefieldMinimapUnitPositionFrame:SetFrameStrata("High")

    -- Hide the origin POI
    local index = 1
    while _G["BattlefieldMinimapPOI"..index] do
        _G["BattlefieldMinimapPOI"..index]:SetAlpha(0)
        index = index + 1
    end

    if not _SVDB.HiddenWhenQuit then BattlefieldMinimap:Show() end

    -- World Map POI
    BattlefieldMinimapPOI = CreateFrame("Frame", "BattlefieldMinimapPOIFrame", BattlefieldMinimap)
    BattlefieldMinimapPOI:SetAllPoints()
    BattlefieldMinimapPOI:SetScale(_SVDB.ScaleModifier / BattlefieldMinimap:GetScale())

    UpdateUnitPositionFrame()

    -- Coords
    BattlefieldMinimapCoords = BattlefieldMinimapScroll:CreateFontString(nil, "OVERLAY", "GameFontNormal")
    BattlefieldMinimapCoords:SetPoint("BOTTOMRIGHT", BattlefieldMinimapScroll, "TOPRIGHT", 0, 2)

    FireSystemEvent("EBFM_INITED")

    RefreshLandmarks()
end

function OnQuit(self)
    if BattlefieldMinimapPOI then
        -- Save settings
        _SVDB.MaskScale         = BattlefieldMinimapScroll:GetWidth() / ORIGIN_WIDTH
        _SVDB.MinimapScale      = BattlefieldMinimap:GetScale()
        _SVDB.HorizontalScroll  = BattlefieldMinimapScroll:GetHorizontalScroll()
        _SVDB.VerticalScroll    = BattlefieldMinimapScroll:GetVerticalScroll()
        _SVDB.HiddenWhenQuit    = not BattlefieldMinimap:IsShown()
    end
end

__SlashCmd__("ebfm", "nocombat", _Locale["on/off - hide during combat"])
function ToggleNoCombat(opt)
    if opt == "on" then
        _SVDB.HideInCombat = true
    elseif opt == "off" then
        _SVDB.HideInCombat = false
    else
        return false
    end
end

__SlashCmd__("ebfm", "lockplayer", _Locale["on/off - lock on the player"])
function ToggleLockPlayer(opt)
    if opt == "on" then
        _SVDB.LockOnPlayer = true
    elseif opt == "off" then
        _SVDB.LockOnPlayer = false
    else
        return false
    end
end

__SlashCmd__("ebfm", "scale", _Locale["1.0 - the scale modifier (0.1-10)"])
function ToggleScaleModifier(scale)
    scale = tonumber(scale)
    if scale and scale >= 0.1 and scale <= 10 then
        _SVDB.ScaleModifier = scale

        BattlefieldMinimapPOI:SetScale(_SVDB.ScaleModifier / BattlefieldMinimap:GetScale())

        UpdateUnitPositionFrame()
        UpdateHitRectInsets()
        RefreshLandmarks()
    else
        return false
    end
end

__AddonSecureHook__ "Blizzard_BattlefieldMinimap"
function BattlefieldMinimapTabDropDown_ShowOpacity()
    OpacityFrame:ClearAllPoints()
    OpacityFrame:SetPoint("TOPRIGHT", BattlefieldMinimapScroll, "TOPLEFT", 0, 7)
end

__AddonSecureHook__ "Blizzard_BattlefieldMinimap"
function BattlefieldMinimap_CreatePOI(index)
    _G["BattlefieldMinimapPOI"..index]:SetAlpha(0)
end

__SystemEvent__() __Thread__()
function PLAYER_REGEN_ENABLED()
    if BattlefieldMinimapScroll and not BattlefieldMinimapScroll:IsVisible() then
        BattlefieldMinimapScroll:SetAlpha(0)
        BattlefieldMinimapScroll:Show()

        local st = GetTime()
        while BattlefieldMinimapScroll:IsVisible() do
            local alpha = (GetTime() - st) / 3.0
            if alpha < 1 then
                BattlefieldMinimapScroll:SetAlpha(alpha)
            else
                BattlefieldMinimapScroll:SetAlpha(1)
                break
            end

            Next()
        end
    end
end

__SystemEvent__()
function PLAYER_REGEN_DISABLED()
    if _SVDB.HideInCombat and BattlefieldMinimapScroll then
        BattlefieldMinimapScroll:Hide()
    end
end

__SystemEvent__ "WORLD_MAP_UPDATE"
    "REQUEST_CEMETERY_LIST_RESPONSE"
    "RESEARCH_ARTIFACT_DIG_SITE_UPDATED"
    "SUPER_TRACKED_QUEST_CHANGED"
    "QUESTLINE_UPDATE"
    "QUEST_LOG_UPDATE"
    "WORLD_QUEST_COMPLETED_BY_SPELL"
    "MINIMAP_UPDATE_TRACKING"
function RefreshLandmarks()
    if not BattlefieldMinimapPOI then return end
    return FireSystemEvent("EBFM_REFRESH")
end

__SecureHook__()
function WorldStateFrame_ToggleBattlefieldMinimap()
    -- Re-show the minimap if blz closed it
    if abs(GetTime() - HiddenTime) < 0.1 then
        BattlefieldMinimap:Show()
    end
end

__SystemEvent__() __Thread__()
function PLAYER_STARTED_MOVING()
    while _SVDB.LockOnPlayer and IsPlayerMoving() and not BattlefieldMinimapScroll:IsMouseOver() and BattlefieldMinimapScroll:IsVisible() do
        local x, y = GetPlayerMapPosition("player")

        if x and y and (x > 0 or y > 0) then
            local scale  = BattlefieldMinimap:GetScale()
            local rx, ry = max(BattlefieldMinimap:GetWidth() - BattlefieldMinimapScroll:GetWidth() / scale, 0), max(BattlefieldMinimap:GetHeight() - BattlefieldMinimapScroll:GetHeight() / scale, 0)

            x = BattlefieldMinimap:GetWidth() * x - BattlefieldMinimapScroll:GetWidth() / 2 / scale
            y = BattlefieldMinimap:GetHeight() * y - BattlefieldMinimapScroll:GetHeight() / 2 / scale

            x = max(0, min(x, rx))
            y = max(0, min(y, ry))

            local cx, cy = BattlefieldMinimapScroll:GetHorizontalScroll(), BattlefieldMinimapScroll:GetVerticalScroll()

            if abs(cx - x) > 2 then if cx > x then x = cx - 2 else x = cx + 2 end end
            if abs(cy - y) > 2 then if cy > y then y = cy - 2 else y = cy + 2 end end

            BattlefieldMinimapScroll:SetHorizontalScroll(x)
            BattlefieldMinimapScroll:SetVerticalScroll(y)

            UpdateHitRectInsets()
        end

        Next()
    end
end

----------------------------------------------
----------------- Addon Helpers --------------
----------------------------------------------
function OnShow(self)
    RefreshLandmarks()
end

function OnHide(self)
    HiddenTime = GetTime()
end

__SystemEvent__()
function EBFM_MOUSE_OVER(cx, cy)
    BattlefieldMinimapCoords:SetText(("(%.2f, %.2f)"):format(cx * 100, cy * 100))
end

__Thread__()
function OnEnter(self)
    ENTER_TASK_ID = ENTER_TASK_ID + 1

    local task = ENTER_TASK_ID

    TabOnEnter()
    UpdateOpacity(0)

    -- Scan the coords
    while task == ENTER_TASK_ID and BattlefieldMinimapScroll:IsMouseOver() do
        local ox, oy = BattlefieldMinimapScroll:GetHorizontalScroll(), BattlefieldMinimapScroll:GetVerticalScroll()
        local cx, cy = GetCursorPosition()
        local rscale = (WorldMapFrame_InWindowedMode() and UIParent or WorldMapFrame):GetScale()
        local scale = BattlefieldMinimap:GetScale()

        cx = (cx / rscale - BattlefieldMinimapScroll:GetLeft()) / scale + ox
        cy = (BattlefieldMinimapScroll:GetTop() - cy / rscale) / scale + oy

        FireSystemEvent("EBFM_MOUSE_OVER", cx / BattlefieldMinimap:GetWidth(), cy / BattlefieldMinimap:GetHeight())

        Next()
    end

    if task == ENTER_TASK_ID then
        BattlefieldMinimapCoords:SetText("")

        if _SVDB.LockOnPlayer and IsPlayerMoving() then PLAYER_STARTED_MOVING() end
        TabOnLeave()

        local tar = BattlefieldMinimapOptions.opacity or 0
        if tar == 0 then return end

        local st = GetTime()

        while task == ENTER_TASK_ID and not BattlefieldMinimapScroll:IsMouseOver() do
            local opacity = (GetTime() - st) / 2.0 * tar

            if opacity < tar then
                UpdateOpacity(opacity)
            else
                UpdateOpacity(tar)
                break
            end

            Next()
        end
    end
end

__Thread__()
function OnMouseDown(self)
    MOUSEDOWN_TIME = GetTime()

    MOVE_MINIMAP_THREADIDX = MOVE_MINIMAP_THREADIDX + 1
    local task = MOVE_MINIMAP_THREADIDX

    local scale  = BattlefieldMinimap:GetScale()
    local rx, ry = max(BattlefieldMinimap:GetWidth() - BattlefieldMinimapScroll:GetWidth() / scale, 0), max(BattlefieldMinimap:GetHeight() - BattlefieldMinimapScroll:GetHeight() / scale, 0)
    local sx, sy = BattlefieldMinimapScroll:GetHorizontalScroll(), BattlefieldMinimapScroll:GetVerticalScroll()
    local cx, cy = GetCursorPosition()

    while task == MOVE_MINIMAP_THREADIDX do
        Next()

        local x, y = GetCursorPosition()

        x = max(0, min(sx - (x - cx) / scale, rx))
        y = max(0, min(sy + (y - cy) / scale, ry))

        BattlefieldMinimapScroll:SetHorizontalScroll(x)
        BattlefieldMinimapScroll:SetVerticalScroll(y)

        UpdateHitRectInsets()
    end
end

__Thread__()
function OnMouseUp(self)
    -- Stop the OnMouseDown thread
    MOVE_MINIMAP_THREADIDX = MOVE_MINIMAP_THREADIDX + 1

    if (GetTime() - MOUSEDOWN_TIME) < 0.15 then
        local task = MOVE_MINIMAP_THREADIDX

        -- Make sure the OnMouseDown thread is finished
        Next() Next()

        if task ~= MOVE_MINIMAP_THREADIDX then return end

        local x, y = GetCursorPosition()
        local rscale = (WorldMapFrame_InWindowedMode() and UIParent or WorldMapFrame):GetScale()

        x = x / rscale - BattlefieldMinimapScroll:GetLeft()
        y = BattlefieldMinimapScroll:GetTop() - y / rscale

        local scale  = BattlefieldMinimap:GetScale()
        local rx, ry = max(BattlefieldMinimap:GetWidth() - BattlefieldMinimapScroll:GetWidth() / scale, 0), max(BattlefieldMinimap:GetHeight() - BattlefieldMinimapScroll:GetHeight() / scale, 0)
        local cx, cy = BattlefieldMinimapScroll:GetHorizontalScroll(), BattlefieldMinimapScroll:GetVerticalScroll()

        x = x / scale + cx - BattlefieldMinimapScroll:GetWidth() / 2 / scale
        y = y / scale + cy - BattlefieldMinimapScroll:GetHeight() / 2 / scale

        x = max(0, min(x, rx))
        y = max(0, min(y, ry))

        while task == MOVE_MINIMAP_THREADIDX and BattlefieldMinimapScroll:IsMouseOver() and abs(cx - x) > 3 or abs(cy - y) > 3 do
            if abs(cx - x) > 3 then if cx > x then cx = cx - 2 else cx = cx + 2 end else cx = x end
            if abs(cy - y) > 3 then if cy > y then cy = cy - 2 else cy = cy + 2 end else cy = y end

            BattlefieldMinimapScroll:SetHorizontalScroll(cx)
            BattlefieldMinimapScroll:SetVerticalScroll(cy)

            UpdateHitRectInsets()

            Next()
        end
    end
end

function OnMouseWheel(self, d)
    if not IsAltKeyDown() then
        -- Scale the minimap
        local ox, oy = BattlefieldMinimapScroll:GetHorizontalScroll(), BattlefieldMinimapScroll:GetVerticalScroll()
        local cx, cy = GetCursorPosition()
        local rscale = (WorldMapFrame_InWindowedMode() and UIParent or WorldMapFrame):GetScale()

        cx = cx / rscale - BattlefieldMinimapScroll:GetLeft()
        cy = BattlefieldMinimapScroll:GetTop() - cy / rscale

        local oscale = self:GetScale()
        local nscale = max(0.1, oscale + d * (IsControlKeyDown() and 0.5 or 0.15))
        local framex = cx / oscale + ox
        local framey = cy / oscale + oy

        self:SetScale(nscale)

        BattlefieldMinimapScroll:SetHorizontalScroll(framex - cx / nscale)
        BattlefieldMinimapScroll:SetVerticalScroll(framey - cy / nscale)
    else
        -- Scale the window
        _SVDB.MaskScale = max(0.3, _SVDB.MaskScale + d * 0.05)
        BattlefieldMinimapScroll:SetSize(ORIGIN_WIDTH * _SVDB.MaskScale, ORIGIN_HEIGHT * _SVDB.MaskScale)
    end

    -- Re-check the scroll ranges
    BattlefieldMinimapScroll:UpdateScrollChildRect()

    local scale  = BattlefieldMinimap:GetScale()
    local rx, ry = max(BattlefieldMinimap:GetWidth() - BattlefieldMinimapScroll:GetWidth() / scale, 0), max(BattlefieldMinimap:GetHeight() - BattlefieldMinimapScroll:GetHeight() / scale, 0)

    if rx <= 0 or ry <= 0 then
        BattlefieldMinimap:SetScale(BattlefieldMinimapScroll:GetWidth() / BattlefieldMinimap:GetWidth())
        BattlefieldMinimapScroll:UpdateScrollChildRect()
        BattlefieldMinimapScroll:SetHorizontalScroll(0)
        BattlefieldMinimapScroll:SetVerticalScroll(0)
    else
        local x, y = BattlefieldMinimapScroll:GetHorizontalScroll(), BattlefieldMinimapScroll:GetVerticalScroll()
        x = max(0, min(x, rx))
        y = max(0, min(y, ry))

        BattlefieldMinimapScroll:SetHorizontalScroll(x)
        BattlefieldMinimapScroll:SetVerticalScroll(y)
    end

    BattlefieldMinimapPOI:SetScale(_SVDB.ScaleModifier / BattlefieldMinimap:GetScale())

    UpdateUnitPositionFrame()
    UpdateHitRectInsets()
    RefreshLandmarks()
end

function TabOnEnter()
    BattlefieldMinimapTabText:Show()
    BattlefieldMinimapTabText:SetAlpha(1)
end

__Thread__()
function TabOnLeave()
    local st = GetTime()

    while not BattlefieldMinimapScroll:IsMouseOver() and not BattlefieldMinimapTab:IsMouseOver() do
        local opacity = (GetTime() - st) / 3.0

        if opacity < 1 then
            BattlefieldMinimapTabText:SetAlpha(1 - opacity)
        else
            BattlefieldMinimapTabText:Hide()
        end

        Next()
    end
end

function UpdateOpacity(opacity)
    local alpha = 1 - opacity
    for i=1, GetNumberOfDetailTiles() do
        _G["BattlefieldMinimap"..i]:SetAlpha(alpha)
    end
    if alpha >= 0.15 then alpha = alpha - 0.15 end
    for i=1, BattlefieldMinimap:GetAttribute("NUM_BATTLEFIELDMAP_OVERLAYS") do
        _G["BattlefieldMinimapOverlay"..i]:SetAlpha(alpha)
    end
end

function UpdateHitRectInsets()
    local x, y = BattlefieldMinimapScroll:GetHorizontalScroll(), BattlefieldMinimapScroll:GetVerticalScroll()
    local scale = BattlefieldMinimap:GetScale()
    local w, h = BattlefieldMinimap:GetSize()
    local mw, mh = BattlefieldMinimapScroll:GetSize()

    BattlefieldMinimap:SetHitRectInsets(x, w - x - mw / scale, y, h - y - mh / scale )
end

function AddPOIButton(poiButton, posX, posY, frameLevelOffset)
    if ( posX and posY ) then
        posX = posX * BattlefieldMinimapPOI:GetWidth()
        posY = posY * BattlefieldMinimapPOI:GetHeight()
        poiButton:SetPoint("CENTER", BattlefieldMinimapPOI, "TOPLEFT", posX, -posY)
        poiButton:SetFrameLevel(poiButton:GetParent():GetFrameLevel() + frameLevelOffset)
    end
end

function UpdateUnitPositionFrame()
    local scale = BattlefieldMinimap:GetScale() / 1.5
    BattlefieldMinimapUnitPositionFrame:SetPinSize("player", 24 / scale)
    BattlefieldMinimapUnitPositionFrame:SetPinSize("party", 8 / scale)
    BattlefieldMinimapUnitPositionFrame:SetPinSize("raid", 8 / scale)

    return BattlefieldMinimapUnitPositionFrame:UpdatePlayerPins()
end
